import { WebGPUError } from '../utils/errors.js'
import { PositiveInteger } from '../utils/index.js'

type SwapChainConfiguration = Omit<GPUCanvasConfiguration, 'device' | 'format'>

export class SwapChain {
  _canvas: HTMLCanvasElement
  _device: GPUDevice
  _context: GPUCanvasContext
  _format: GPUTextureFormat
  _width: number
  _height: number

  _configuration: SwapChainConfiguration

  get context() {
    return this._context
  }

  get format() {
    return this._format
  }

  get width() {
    return this._width
  }

  get height() {
    return this._height
  }

  constructor(canvas: HTMLCanvasElement, device: GPUDevice, context?: GPUCanvasContext) {
    this._canvas = canvas
    this._device = device

    this._context = context || canvas.getContext('webgpu')

    if (!this._context) {
      throw WebGPUError.contextFailure()
    }

    this._format = navigator.gpu.getPreferredCanvasFormat()
    this._width = canvas.width
    this._height = canvas.height
  }

  configure(configuration?: SwapChainConfiguration) {
    if (configuration) {
      this._configuration = configuration
    }

    this._width = this._canvas.width
    this._height = this._canvas.height

    this._context.configure({
      device: this._device,
      format: this._format,
      alphaMode: 'opaque',
      ...this._configuration
    })
  }

  getCurrentTextureView() {
    return this._context.getCurrentTexture().createView()
  }

  resize<const T extends number>(width: PositiveInteger<T>, height: PositiveInteger<T>) {
    if (width <= 0 || height <= 0) {
      return
    }

    if (this._width !== width || this._height !== height) {
      this._canvas.width = width
      this._canvas.height = height
      this.configure()
    }
  }
}

