import { BindGroupLayout } from './bind-group-layout.js'
import { ShaderPair, ShaderModule } from './shader-module.js'
import { MaterialError, WebGPUObjectError } from '../utils/errors.js'
import { ResourceType } from 'wgsl_reflect'
import { BindGroup } from './bind-group.js'

/** @import { FragmentStateDescriptor, VertexStateDescriptor } from './shader-module.js' */

/**
 * @typedef ShaderPairDescriptor
 * @property {ShaderModule} vertex
 * @property {ShaderModule} [fragment]
 */

/**
 * @typedef {
     ShaderPairDescriptor &
     { bindGroupLayouts?: BindGroupLayout[] }
 * } MaterialDescriptor
 */

export class Material {
  _device
  _shaders
  _bindGroupLayouts
  _pipelineLayout

  get shaders() {
    return this._shaders
  }

  get bindGroupLayouts() {
    return this._bindGroupLayouts
  }

  /**
   * @param {GPUDevice} device
   * @param {MaterialDescriptor} descriptor
   */
  constructor(device, descriptor) {
    this._device = device
    this._shaders = Material._reflectShaders(descriptor)
    const bgl = descriptor.bindGroupLayouts

    if (bgl && bgl.length > 0) {
      this._bindGroupLayouts = bgl
    } else {
      this._bindGroupLayouts = this._reflectBindGroupLayouts(device, this._shaders)
    }

    if (this._bindGroupLayouts && this.bindGroupLayouts.length > 0) {
      try {
        this._pipelineLayout = device.createPipelineLayout({
          bindGroupLayouts: this._bindGroupLayouts.map(bgl => bgl.handle)
        })
      } catch (err) {
        throw WebGPUObjectError.from(err, Material)
      }
    }
  }
  /**
   * Attempts to handle shader modules which represent multiple
   * shader types.
   *
   * @param {ShaderPairDescriptor} shaders
   * @returns {ShaderPair}
   */
  static _reflectShaders(shaders) {
    if (shaders == null) {
      throw MaterialError.missingShader('vertex')
    }

    if ('vertex' in shaders) {
      return ShaderPair.fromPair(shaders)
    }
  }

  /**
   * @param {GPUDevice} device
   * @param {ShaderPair} shaders 
   * @returns {BindGroupLayout[]}
   */
  _reflectBindGroupLayouts(device, shaders) {
    const layouts = shaders.createBindGroupLayoutEntries()
    return layouts.map(entries => BindGroupLayout.create(device, { entries }))
  }

  /**
   *
   */
  createBindGroup(groupIndex, resources, label) {
    if (groupIndex < 0 || groupIndex >= this._bindGroupLayouts.length) {
      throw new Error(`Invalid bind group index: ${groupIndex}`)
    }

    const bgl = this._bindGroupLayouts[groupIndex]

    let entries = []
    for (const name in resources) {
      const resource = resources[name]

      const variableInfo = this._shaders.findVariableInfo(name, groupIndex)

      const entry = {
        binding: variableInfo.binding,
        resource: resource.toBindingResource()
      }

      if (variableInfo.resourceType === ResourceType.Uniform || variableInfo.resourceType === ResourceType.Storage) {
        // TODO: handle user provided offset/size
      }

      entries.push(entry)
    }

    entries.sort((a, b) => a.binding - b.binding)

    return BindGroup.create(this._device, {
      layout: bgl.handle,
      entries,
      label
    })
  }

  /**
   * @typedef MaterialPipelineDescriptor
   * @property {string} [label]
   * @property {GPUPipelineLayout} [pipelineLayout]
   * @property {VertexStateDescriptor} vertex
   * @property {FragmentStateDescriptor} [fragment]
   * @property {GPUPrimitiveState} [primitive]
   */

  /**
   * @param {MaterialPipelineDescriptor} descriptor
   * @returns {GPURenderPipelineDescriptor}
   */
  getRenderPipelineDescriptor(descriptor) {
    const { fragment, vertex } = this.shaders.getRenderPipelineStates(descriptor)

    return {
      label: descriptor.label,
      layout: descriptor.pipelineLayout || this._pipelineLayout,
      fragment,
      vertex,
      primitive: descriptor.primitive || { topology: 'triangle-list' }
    }
  }
}
