import { ResourceType, VariableInfo, WgslReflect } from 'wgsl_reflect'
import { WebGPUObjectError } from '../utils/errors.js'
import { BindingMap, GroupBindingMap, VariableStageInfo } from '../utils/bindings.js'

import {
  accessToBufferType,
  accessToStorageTextureAccess,
  parseTextureType,
  typeToSamplerBindingType,
  typeToTextureSampleType,
  typeToViewDimension,
  wgslToWgpuFormat
} from '../utils/wgsl-to-wgpu.js'
import { BufferBindingType } from '../enum.js'

/** @import { WGSLAccess, WGSLSamplerType } from '../utils/wgsl-to-wgpu.js' */

export class ShaderModule {
  _handle: GPUShaderModule
  _code: string

  _reflection: WgslReflect | undefined

  get handle() {
    return this._handle
  }

  get label() {
    return this._handle.label
  }

  constructor(device: GPUDevice, descriptor: GPUShaderModuleDescriptor) {
    this._code = descriptor.code

    try {
      this._handle = device.createShaderModule(descriptor)
    } catch (err) {
      throw WebGPUObjectError.from(err, ShaderModule)
    }
  }

  static create(device: GPUDevice, descriptor: GPUShaderModuleDescriptor) {
    return new ShaderModule(device, descriptor)
  }

  reflect() {
    if (this._reflection == null) {
      this._reflection = new WgslReflect(this._code)
      // no longer needed allow the GC to collect it
      this._code = undefined
    }

    return this._reflection
  }
}

export interface FragmentStateDescriptor {
  constants?: Record<string, GPUPipelineConstantValue>
  targets?: GPUColorTargetState[]
}

export interface VertexStateDescriptor {
  constants?: Record<string, GPUPipelineConstantValue>
  buffers?: GPUVertexBufferLayout[]
}

export interface ShaderPairStateDescriptor {
  fragment?: FragmentStateDescriptor
  vertex: VertexStateDescriptor
}

export class ReflectedShader {
  static _reflectTypes = ['uniforms', 'storage', 'textures', 'samplers']
  _module: ShaderModule

  get module() {
    return this._module
  }


  constructor(shader: ShaderModule) {
    this._module = shader
  }


  findVariableInfo(name: string, group: number): VariableInfo | undefined {
    const reflection = this.module.reflect()
    for (const type of ReflectedShader._reflectTypes) {
      const variables = reflection[type]
      if (variables) {
        const variable = variables.find((v: VariableInfo) => v.name === name && v.group === group)
        if (variable) {
          return variable
        }
      }
    }
  }

  getEntrypoint(stageName: string): string | undefined {
    const entry = this.module.reflect().entry

    // TODO: determine how to correctly handle
    // multiple entrypoints to the same stage
    return entry[stageName].length === 1 ?
      entry[stageName][0].name : undefined
  }

  getShaderStages(): GPUShaderStageFlags {
    const entry = this._module.reflect().entry
    let stages = 0

    stages |= entry.vertex.length > 0 ?
      GPUShaderStage.VERTEX : 0

    stages |= entry.fragment.length > 0 ?
      GPUShaderStage.FRAGMENT : 0

    stages |= entry.compute.length > 0 ?
      GPUShaderStage.COMPUTE : 0

    return stages
  }

  hasStage(stages: GPUShaderStageFlags) {
    return this.getShaderStages()
      & stages
  }

  getBindingsForStage(stages: GPUShaderStageFlags, out: GroupBindingMap = new GroupBindingMap()) {
    const groups = this._module.reflect().getBindGroups()

    groups.forEach((bindings, groupIndex) => {
      if (!out.has(groupIndex)) {
        out.set(groupIndex, new BindingMap())
      }

      const bindingsMap = out.get(groupIndex)

      bindings.filter(x => x != null).forEach(variableInfo => {
        if (!bindingsMap.has(variableInfo.binding)) {
          bindingsMap.set(variableInfo.binding, { variableInfo, stages })
        } else {
          bindingsMap.get(variableInfo.binding).stages |= stages
        }
      })
    })

    return out
  }

  static _sortKeyIndices(map: Map<any, any>): number[] {
    return Array.from(map.keys()).sort((a, b) => a - b)
  }

  static _parseUniform(_variableInfo: VariableInfo): GPUBufferBindingLayout {
    return {
      type: BufferBindingType.Uniform,
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  static _parseStorage(variableInfo: VariableInfo): GPUBufferBindingLayout {
    return {
      type: accessToBufferType(
        variableInfo.access
      ),
      // TODO: infer these two properties
      hasDynamicOffset: false,
      minBindingSize: 0
    }
  }

  static _parseTexture(variableInfo: VariableInfo): GPUTextureBindingLayout {
    const [type, sampledType] = parseTextureType(
      variableInfo.type.name
    )

    return {
      sampleType: typeToTextureSampleType(type, sampledType),
      viewDimension: typeToViewDimension(type),
      multisampled: type.includes('multisampled')
    }
  }

  static _parseSampler(variableInfo: VariableInfo): GPUSamplerBindingLayout {
    return {
      type: typeToSamplerBindingType(
        variableInfo.type.name
      )
    }
  }

  static _parseStorageTexture(variableInfo: VariableInfo): GPUStorageTextureBindingLayout {
    const [type] = parseTextureType(variableInfo.type.name)

    return {
      access: accessToStorageTextureAccess(
        variableInfo.access
      ),
      format: wgslToWgpuFormat(variableInfo.type.name),
      viewDimension: typeToViewDimension(type)
    }
  }

  static _variableInfoToEntry(variableStageInfo: VariableStageInfo): GPUBindGroupLayoutEntry {
    const { stages: visibility, variableInfo } = variableStageInfo

    switch (variableInfo.resourceType) {
      case ResourceType.Uniform:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseUniform(variableInfo)
        }
      case ResourceType.Storage:
        return {
          binding: variableInfo.binding,
          visibility,
          buffer: this._parseStorage(variableInfo)
        }
      case ResourceType.Texture:
        return {
          binding: variableInfo.binding,
          visibility,
          texture: this._parseTexture(variableInfo)
        }
      case ResourceType.Sampler:
        return {
          binding: variableInfo.binding,
          visibility,
          sampler: this._parseSampler(variableInfo)
        }
      case ResourceType.StorageTexture:
        return {
          binding: variableInfo.binding,
          visibility,
          storageTexture: this._parseStorageTexture(variableInfo)
        }
      default:
        console.warn(`Unsupported resource type for reflection: ${ResourceType[variableInfo.resourceType]}`)
        return
    }
  }

  static createBindGroupLayoutEntries(groupBindings: GroupBindingMap) {
    const sortedGroupIndices = this._sortKeyIndices(groupBindings)

    return sortedGroupIndices.map(groupIndex => {
      const bindingsMap = groupBindings.get(groupIndex)
      return this._sortKeyIndices(bindingsMap).map(bindingIndex => (
        this._variableInfoToEntry(bindingsMap.get(bindingIndex))
      ))
    })
      .filter(desc => desc.length > 0)
  }
}

export class ShaderPair {
  _vertex: ReflectedShader
  _fragment: ReflectedShader

  constructor(vertex: ReflectedShader, fragment?: ReflectedShader) {
    if (!vertex) {
      throw new Error('Missing vertex shader')
    }

    if (!vertex.hasStage(GPUShaderStage.VERTEX)) {
      throw new Error('Vertex shader does not have a vertex entrypoint.')
    }

    this._vertex = vertex

    if (fragment) {
      if (fragment.hasStage(GPUShaderStage.FRAGMENT)) {
        this._fragment = fragment
      } else {
        throw new Error('Fragment shader does not have a fragment entrypoint.')
      }
    } else if (this._vertex.hasStage(GPUShaderStage.FRAGMENT)) {
      // this is a unified shader, use the vertex as the fragment
      this._fragment = vertex
    } else {
      throw new Error('Missing fragment shader.')
    }
  }

  static fromUnifiedShader(shader: ShaderModule) {
    return new ShaderPair(
      new ReflectedShader(shader)
    )
  }

  static fromPair(value: {
    vertex: ShaderModule
    fragment?: ShaderModule
  }) {
    const vert = new ReflectedShader(value.vertex)
    const frag = value.fragment && new ReflectedShader(value.fragment)
    return new ShaderPair(vert, frag)
  }

  _createGroupBindings() {
    const groupBindings = new GroupBindingMap()

    this._vertex.getBindingsForStage(
      GPUShaderStage.VERTEX,
      groupBindings
    )

    this._fragment.getBindingsForStage(
      GPUShaderStage.FRAGMENT,
      groupBindings
    )

    return groupBindings
  }

  createBindGroupLayoutEntries() {
    return ReflectedShader.createBindGroupLayoutEntries(
      this._createGroupBindings()
    )
  }

  _getFragmentState(descriptor: FragmentStateDescriptor): GPUFragmentState {
    return {
      module: this._fragment.module.handle,
      entryPoint: this._fragment.getEntrypoint('fragment'),
      constants: descriptor.constants || {},
      targets: descriptor.targets || []
    }
  }

  _getVertexState(descriptor: VertexStateDescriptor): GPUVertexState {
    return {
      module: this._vertex.module.handle,
      entryPoint: this._vertex.getEntrypoint('vertex'),
      constants: descriptor.constants || {},
      buffers: descriptor.buffers || []
    }
  }

  findVariableInfo(name: string, group: number): VariableInfo | undefined {
    let variableInfo = this._vertex.findVariableInfo(name, group)

    if (!variableInfo && this._fragment !== this._vertex) {
      variableInfo = this._fragment.findVariableInfo(name, group)
    }

    return variableInfo
  }

  getRenderPipelineStates(descriptor: ShaderPairStateDescriptor): Pick<GPURenderPipelineDescriptor, 'vertex' | 'fragment'> {
    return {
      fragment: this._getFragmentState(descriptor.fragment),
      vertex: this._getVertexState(descriptor.vertex),
    }
  }
}

