import { CommandRecorder } from '../core/command-recorder.js'
import { BitFlags } from '../utils/bitflags.js'
import { WebGPUObjectError } from '../utils/errors.js'
import { ResourceType } from '../utils/internal-enums.js'
import { textureToImageDimension } from '../utils/wgsl-to-wgpu.js'

/** @import { BindGroupEntry } from '../core/graphics-device.js' */

export class Texture {
  static _defaultUsage = GPUTextureUsage.TEXTURE_BINDING
    | GPUTextureUsage.COPY_DST
    | GPUTextureUsage.RENDER_ATTACHMENT

  _device
  _handle

  /** @type {GPUTextureView | undefined} */
  _defaultView

  get handle() {
    return this._handle
  }

  get width() {
    return this._handle.width
  }

  get height() {
    return this._handle.height
  }

  get format() {
    return this._handle.format
  }

  get depthOrArrayLayers() {
    return this._handle.depthOrArrayLayers
  }

  get usage() {
    return this._handle.usage
  }

  get dimension() {
    return this._handle.dimension
  }

  get mipLevelCount() {
    return this._handle.mipLevelCount
  }

  get label() {
    return this._handle.label
  }

  get resourceType() {
    return ResourceType.TextureView
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUTexture} texture
   */
  constructor(device, texture) {
    this._device = device
    this._handle = texture
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUTextureDescriptor} descriptor
   */
  static create(device, descriptor) {
    try {
      return new Texture(
        device,
        device.createTexture(descriptor)
      )
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  static _generateMipLevels(size) {
    const max = Math.max.apply(undefined, size)
    return 1 + Math.log2(max) | 0
  }


  /**
   * @param {GPUDevice} device
   * @param {string | URL} url
   * @param {GPUTextureDescriptor} desciptor
   */
  static async fromUrl(device, url, descriptor) {
    try {
      const response = await fetch(url)

      if (!response.ok) {
        throw new Error(`Failed to fetch remote resource: ${response.statusText}`)
      }

      const usage = options.usage || Texture._defaultUsage
      const dimension = descriptor.dimension ? textureToImageDimension(descriptor.dimension) : '2d'

      const blob = await response.blob()
      const bitmap = await createImageBitmap(blob)
      const size = [bitmap.width, bitmap.height]

      const desc = {
        usage,
        dimension,
        size,
        format: descriptor.format || 'rgba8unorm',
        mipLevelCount: descriptor.mipLevelCount || Texture._generateMipCount(...size),
        ...descriptor,
      }

      const texture = Texture.create(device, desc)

      texture.upload(bitmap)

      return texture
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  /**
   * @param {GPUDevice} device
   * @param {GPUExtent3DStrict} size
   * @param {GPUTextureFormat} format
   * @param {GPUTextureDescriptor} descriptor
   */
  static createRenderTarget(device, size, format, descriptor) {
    const usage = descriptor.usage || Texture._defaultUsage

    return Texture.create(device, {
      size,
      format,
      usage,
      ...descriptor
    })
  }

  /**
   * @typedef UploadTextureInfo
   * @property {GPUTexelCopyTextureInfo} destination
   * @property {GPUTexelCopyBufferLayout} dataLayout
   * @property {GPUExtent3DStrict} size
   */

  /**
   * @param {GPUAllowSharedBufferSource} source
   * @param {UploadTextureInfo} [options={}] 
   */
  upload(source, options = {}) {
    const mipLevel = options.destination.mipLevel || 0
    const size = options.size || [
      Math.max(1, this.width >> mipLevel),
      Math.max(1, this.height >> mipLevel),
      this.depthOrArrayLayers
    ]

    try {
      this._device.queue.writeTexture(
        { ...options.destination, texture: this._handle },
        source,
        options.dataLayout,
        size
      )
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  /**
    * @param {GPUCommandEncoder} commandEncoder
   */
  generateMipmaps(commandEncoder) {
    const requiredUsage = GPUTextureUsage.COPY_SRC
      | GPUTextureUsage.COPY_DST
      | GPUTextureUsage.RENDER_ATTACHMENT

    if (!BitFlags.has(this.usage & requiredUsage)) {
      throw new Error('Texture does not have the required usage flags for mipmap generation')
    }

    for (let i = 1; i < this.mipLevelCount; ++i) {

    }
  }

  /**
   * @param {GPUTextureViewDescriptor} [descriptor]
   * @throws {TextureError}
   */
  createDefaultView(descriptor) {
    if (!descriptor && this._defaultView) {
      return this._defaultView
    }

    try {
      const view = this._handle.createView(descriptor)

      if (!descriptor) {
        this._defaultView = view
      }

      return view
    } catch (err) {
      throw WebGPUObjectError.from(err, Texture)
    }
  }

  getView() {
    return this.createDefaultView()
  }

  toBindingResource() {
    return this.getView()
  }

  destroy() {
    this._handle?.destroy()
    this._handle = undefined
    this._defaultView = undefined
  }
}

