import { Newable } from './index.js'

export class WebGPUError extends Error {
  constructor(message: string, options?: ErrorOptions) {
    super(`WebGPUError: ${message}`, options)
  }

  static unsupported() {
    return new WebGPUError('WebGPU is not supported on this browser')
  }

  static contextFailure() {
    return new WebGPUError('Failed to request WebGPU context')
  }

  static adapterUnavailable() {
    return new WebGPUError('Failed to request GPU adapter')
  }

  static deviceUnavailable() {
    return new WebGPUError('Failed to request GPU device')
  }
}

export class GraphicsDeviceError extends Error {
  static uninitialized() {
    return new GraphicsDeviceError('GraphicsDeviceError: GraphicsDevice not initialized')
  }
}

export class CommandRecorderError extends Error {
  static activeRenderPass() {
    return new CommandRecorderError('CommandRecorderError: can\'t begin new render pass while another is active. call CommandRecorder.end()')
  }
}

export class WebGPUObjectError extends Error {
  static from<T extends Error>(cause: T, type: string | Newable) {
    const name = typeof type === 'string' ? type : type.name
    return new WebGPUObjectError(`could not create ${name}`, { cause })
  }
}

export class BufferError extends WebGPUObjectError {
  constructor(message: string, options?: ErrorOptions) {
    super(`BufferError: ${message}`, options)
  }

  static from<T extends Error>(cause: T) {
    return new BufferError('could not create buffer', { cause })
  }

  static invalidRead() {
    return new BufferError('cannot read a buffer without MAP_READ usage')
  }

  static outOfBounds(offset: number, size: number) {
    return new BufferError(`buffer offset/size (${offset}/${size}) exceeds buffer dimensions`)
  }
}

export class MaterialError extends WebGPUObjectError {
  constructor(message: string, options?: ErrorOptions) {
    super(`MaterialError: ${message}`, options)
  }

  static from<T extends Error>(cause: T) {
    return new BufferError('could not create material', { cause })
  }

  static missingShader(shaderType: string) {
    return new BufferError(`missing ${shaderType} shader`)
  }
}
